open Spoc.Vector

type point = {
	mutable x: float;
	mutable y: float;
}

external getsizeofpoint: unit -> int = "custom_getsizeofpoint"
external extget: customarray -> int -> point = "custom_extget"
external extset: customarray -> int -> point -> unit = "custom_extset"

let customPoint = {
	elt = { x = 0.; y = 0.};
	size = getsizeofpoint();
	get = extget;
	set = extset;
}

let nbPoint = 2_000_000;;
let ray = 10.0;;

 kernel gpuPi : point Spoc.Vector.vcustom -> Spoc.Vector.vfloat32 -> int  -> float -> unit = "kernels/CustomPi" "pi"
 


let cpuPI field =
	let resVec = Array.map (fun pt -> (pt.x *. pt.x +. pt.y *. pt.y <= (ray*.ray))) field in
	let res = Array.fold_left (fun i res -> if res then i + 1 else i) 0 resVec
	in res

let gpuPI gpuField vbool devices =
	let threadsPerBlock = 512 in
	let blocksPerGrid = (nbPoint / 32 + threadsPerBlock - 1) / threadsPerBlock in
	let block = {Spoc.Kernel.blockX = threadsPerBlock; Spoc.Kernel.blockY = 1; Spoc.Kernel.blockZ = 1} 
	and grid = {Spoc.Kernel.gridX = blocksPerGrid; Spoc.Kernel.gridY = 1; Spoc.Kernel.gridZ = 1} in
	Spoc.Kernel.run devices.(0) (block,grid) ~cached: true ~queue_id:(0) (gpuPi gpuField vbool (nbPoint) (ray*.ray));
	let pio4 = int_of_float (Cublas.run devices.(0) (Cublas.cublasSasum (nbPoint) vbool 1))
	in
		pio4
	
let multiGpuPI gpuField1 gpuField2 vbool1 vbool2 devices =
	let threadsPerBlock = 512 in
	let blocksPerGrid = (nbPoint / 32 / 2 + threadsPerBlock - 1) / threadsPerBlock in
	let block = {Spoc.Kernel.blockX = threadsPerBlock; Spoc.Kernel.blockY = 1; Spoc.Kernel.blockZ = 1} 
	and grid = {Spoc.Kernel.gridX = blocksPerGrid; Spoc.Kernel.gridY = 1; Spoc.Kernel.gridZ = 1} in
	
	Spoc.Kernel.run devices.(0) (block,grid) ~cached: true ~queue_id:(0) (gpuPi gpuField1 vbool1 (nbPoint / 2) (ray*.ray));
	Spoc.Kernel.run devices.(2) (block,grid) ~cached: true ~queue_id:(0) (gpuPi gpuField2 vbool2 (nbPoint / 2) (ray*.ray));
	let pio4_1 = int_of_float (Cublas.run devices.(0) (Cublas.cublasSasum (nbPoint / 2) vbool1 1))
	and pio4_2 = int_of_float (Cublas.run devices.(2) (Cublas.cublasSasum (nbPoint / 2) vbool2 1)) in
	let pio4 = pio4_1 + pio4_2
	in
	pio4

let _ =
	Random.self_init();
	let pio4 = ref 0 in

	(*CPU COMPUTATION*)
	let t0 = (Unix.gettimeofday ()) in
	let field = Array.init nbPoint (fun _ -> { x = Random.float ray; y = Random.float ray }) in
	let t1 = (Unix.gettimeofday ()) in
			pio4 :=	cpuPI field;
	Printf.printf "%d/%d = %.10G\n" !pio4 nbPoint (4.*.(float !pio4)/.(float nbPoint));

	(*GPU COMPUTATION*)
	let t2 = (Unix.gettimeofday ()) in
	let devices = Spoc.Devices.init () in
		Cublas.init ();
		Spoc.Mem.auto_transfers false;
		let vbool = Spoc.Vector.create (float32) (nbPoint) in
		let gpuField = Spoc.Tools.map (fun  _ -> { x = Random.float ray; y = Random.float ray }) (Custom customPoint) vbool in
		Spoc.Mem.to_device gpuField ~queue_id: (1) devices.(0);
		Spoc.Mem.to_device vbool ~queue_id: (0) devices.(0);
		let t3 = (Unix.gettimeofday ()) in
				pio4 :=	gpuPI gpuField vbool devices;
			Printf.printf "%d/%d = %.10G\n" !pio4 nbPoint (4.*.(float !pio4)/.(float nbPoint) );
			Pervasives.flush stdout;
	Cublas.shutdown ();;