open Spoc

exception NOT_INITIALIZED
exception ALLOC_FAILED
exception INVALID_VALUE
exception ARCH_MISMATCH
exception MAPPING_ERROR
exception EXECUTION_FAILED
exception INTERNAL_ERROR
exception UNKNOWN_ERROR

let _ = 
	Callback.register_exception "cublas_status_not_initialised" NOT_INITIALIZED;
	Callback.register_exception "cublas_status_alloc_failed" ALLOC_FAILED;
	Callback.register_exception "cublas_status_invalid_value" INVALID_VALUE;
	Callback.register_exception "cublas_status_arch_mismatch" ARCH_MISMATCH;
	Callback.register_exception "cublas_status_mapping_error" MAPPING_ERROR;
	Callback.register_exception "cublas_status_execution_failed" EXECUTION_FAILED;
	Callback.register_exception "cublas_status_internal_error" INTERNAL_ERROR;
	Callback.register_exception "cublas_error_unknown" UNKNOWN_ERROR




external init: unit -> unit = "spoc_cublasInit"
external shutdown: unit -> unit = "spoc_cublasShutdown"
external getError: unit -> unit = "spoc_cublasGetError"

(* Blas1 *)
external cublasIsamax: int -> Vector.vfloat32 -> int -> Devices.device ->  int = "spoc_cublasIsamax_b" "spoc_cublasIsamax_n"
external cublasIsamin: int -> Vector.vfloat32 -> int -> Devices.device ->  int = "spoc_cublasIsamin_b" "spoc_cublasIsamin_n"
external cublasSasum: int -> Vector.vfloat32 -> int -> Devices.device -> float = "spoc_cublasSasum_b" "spoc_cublasSasum_n"
external cublasSaxpy: int -> float -> Vector.vfloat32 -> int -> Vector.vfloat32 -> int -> Devices.device -> unit = "spoc_cublasSaxpy_b" "spoc_cublasSaxpy_n"
external cublasScopy: int -> Vector.vfloat32 -> int -> Vector.vfloat32  -> int -> Devices.device -> unit = "spoc_cublasScopy_b" "spoc_cublasScopy_n"
external cublasSdot: int -> Vector.vfloat32 -> int -> Vector.vfloat32  -> int -> Devices.device -> float = "spoc_cublasSdot_b" "spoc_cublasSdot_n"
external cublasSnrm2: int -> Vector.vfloat32 -> int -> Devices.device -> float = "spoc_cublasSnrm2_b" "spoc_cublasSnrm2_n"
external cublasSrot: int -> Vector.vfloat32 -> int -> Vector.vfloat32  -> int -> float -> float -> Devices.device -> unit = "spoc_cublasSrot_b" "spoc_cublasSrot_n"
external cublasSrotg: Vector.vfloat32 -> Vector.vfloat32 -> Vector.vfloat32 -> Vector.vfloat32 ->  unit = "spoc_cublasSrotg_b" "spoc_cublasSrotg_n"
external cublasSrotm: int -> Vector.vfloat32 -> int -> Vector.vfloat32  -> int -> Vector.vfloat32 -> Devices.device -> unit = "spoc_cublasSrotm_b" "spoc_cublasSrotm_n"
external cublasSrotmg: Vector.vfloat32 -> Vector.vfloat32 -> Vector.vfloat32 -> Vector.vfloat32 -> Vector.vfloat32 -> unit = "spoc_cublasSrotmg_b" "spoc_cublasSrotmg_n"
external cublasSscal: int -> float -> Vector.vfloat32 -> int -> Devices.device -> unit = "spoc_cublasSscal_b" "spoc_cublasSscal_n"
external cublasSswap: int -> Vector.vfloat32 -> int -> Vector.vfloat32  -> int -> Devices.device -> unit = "spoc_cublasSswap_b" "spoc_cublasSswap_n"

external cublasCaxpy: int -> Complex.t -> Vector.vcomplex32 -> int -> Vector.vcomplex32 -> int -> Devices.device -> unit = "spoc_cublasCaxpy_b" "spoc_cublasCaxpy_n"
external cublasScasum: int -> Vector.vcomplex32 -> int -> Devices.device -> float = "spoc_cublasScasum_b" "spoc_cublasScasum_n"





let check_auto vecs dev= 
	if !Mem.auto then
		for i = 0 to Array.length vecs - 1 do
			match vecs.(i).Vector.dev with
				| Vector.No_dev  -> Mem.to_device vecs.(i) dev
				| Vector.Dev d  -> 
					if vecs.(i).Vector.device != d.Devices.general_info.Devices.id then
						Mem.to_device vecs.(i) dev
		done
		
(** Blas1 *)
(** Single-Precision Blas1 functions *)


let cublasIsamax n x incx dev =
	check_auto [|x|] dev;
	cublasIsamax n x incx dev

let cublasIsamin n x incx dev =
	check_auto [|x|] dev;
	cublasIsamin n x incx dev

let cublasSasum n x incx dev =
	check_auto [|x|] dev;
	cublasSasum n x incx dev

let cublasSaxpy n alpha x incx y incy dev =
	check_auto [|x; y|] dev;
	cublasSaxpy n alpha x incx y incy dev


let cublasScopy n x incx y incy dev =
	check_auto [|x; y|] dev;
	cublasScopy n x incx y incy dev

let cublasSdot n x incx y incy dev =
	check_auto [|x; y|] dev;
	cublasSdot n x incx y incy dev

let cublasSnrm2 n x incx dev =
	check_auto [|x|] dev;
	cublasSnrm2 n x incx dev

let cublasSrot n x incx y incy sc ss dev =
	check_auto [|x; y|] dev;
	cublasSrot n x incx y incy sc ss dev

let cublasSrotg host_sa host_sb host_sc host_ss dev =
	let vecs = [|host_sa; host_sb; host_sc; host_ss|]
	in 
	if !Mem.auto then
	begin
		for i = 0 to Array.length vecs - 1 do
			if vecs.(i).Vector.device != -1 then
				Mem.to_cpu vecs.(i) ()
		done;
	end;
	cublasSrotg host_sa host_sb host_sc host_ss

let cublasSrotm n x incx y incy sparam dev =
	check_auto [|x; y; sparam|] dev;
	cublasSrotm n x incx y incy sparam dev

let cublasSrotmg host_psd1 host_psd2 host_psx1 host_psy1 host_sparam dev =
	let vecs = [|host_psd1; host_psd2; host_psx1; host_psy1; host_sparam|]
	in 
	if !Mem.auto then
	begin
		for i = 0 to Array.length vecs - 1 do
			if vecs.(i).Vector.device != -1 then
				Mem.to_cpu vecs.(i) ()
		done;
	end;
	cublasSrotmg host_psd1 host_psd2 host_psx1 host_psy1 host_sparam


let cublasSscal n alpha x incx dev =
	check_auto [|x|] dev;
	cublasSscal n alpha x incx dev
	
let cublasSswap n x incx y incy dev =
	check_auto [|x; y|] dev;
	cublasSswap n x incx y incy dev	

(** Single-Precision Complex Blas1 functions *)


let cublasCaxpy n alpha x incx y incy dev =
	check_auto [|x; y|] dev;
	cublasCaxpy n alpha x incx y incy dev	

let cublasScasum n x incx dev =
	check_auto [|x|] dev;
	cublasScasum n x incx dev

(*Blas3*)


(** Blas3 *)
(** Single-Precision Blas3 functions *)			

external cublasSgemm: char -> char -> int -> int -> int -> float -> Vector.vfloat32 -> int -> Vector.vfloat32 -> int -> float -> Vector.vfloat32 -> int -> Devices.device -> unit = "spoc_cublasSgemm_b" "spoc_cublasSgemm_n"																		

let cublasSgemm transa transb m n k alpha a lda b ldb beta c ldc dev =
	check_auto [|a; b; c|] dev;
	cublasSgemm transa transb m n k alpha a lda b ldb beta c ldc dev
	
																									
let run dev f =
	f dev 
