type dim3 = {x:int; y:int; z:int}


external cuda_init : unit -> unit = "spoc_cuInit" 
external cl_init : unit -> unit = "spoc_clInit" 


type specificLibrary =
	| Cuda
	| OpenCL
	| Both

type context

type generalInfo = {
	name:string;
	totalGlobalMem:int;
	localMemSize:int; (*shared mem per block in cuda*)	
	clockRate:int;
	totalConstMem:int; (*MAX_CONSTANT_BUFFER_SIZE in OpenCL*)
	multiProcessorCount:int; (*MAX_COMPUTE_UNIT in OpenCL*)
	eccEnabled:bool; (*ERROR_CORRECTION_SUPPORT in OpenCL*)
	specific : specificLibrary;
	id:int;
	ctx:context
	}

type cudaInfo = {	
	major:int;
  minor:int;
  regsPerBlock:int;
	warpSize:int;
	memPitch:int;
	maxThreadsPerBlock:int;
  maxThreadsDim:dim3;
	maxGridSize:dim3;
	textureAlignment:int;
	deviceOverlap:bool;
	kernelExecTimeoutEnabled:bool;
	integrated:bool;
	canMapHostMemory:bool;
	computeMode:int;
	concurrentKernels:bool;
	pciBusID:int;
  pciDeviceID:int;
	}	

type platformInfo = {
	platform_profile: string;
	platform_version  : string; 
	platform_name : string;
	platform_vendor : string;
	platform_extensions : string;
	num_devices : int;
	}

type deviceType = 
	CL_DEVICE_TYPE_CPU | CL_DEVICE_TYPE_GPU | CL_DEVICE_TYPE_ACCELERATOR | CL_DEVICE_TYPE_DEFAULT

type clDeviceFPConfig = 
	| CL_FP_DENORM
	| CL_FP_INF_NAN
	| CL_FP_ROUND_TO_NEAREST
	| CL_FP_ROUND_TO_ZERO
	| CL_FP_ROUND_TO_INF
	| CL_FP_FMA
	| CL_FP_NONE



type clDeviceQueueProperties =
	| CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE | CL_QUEUE_PROFILING_ENABLE

type clDeviceGlobalMemCacheType = 
	| CL_READ_WRITE_CACHE	
	| CL_READ_ONLY_CACHE 
	| CL_NONE 

type clDeviceLocalMemType = 
	| CL_LOCAL | CL_GLOBAL

type clDeviceExecutionCapabilities = 
	| CL_EXEC_KERNEL | CL_EXEC_NATIVE_KERNEL

type clDeviceID

type openCLInfo = {
		platform_info : platformInfo;
		device_type : deviceType;
		profile : string;
		version : string;
		vendor : string;
		extensions : string;
		vendor_id : int;
		max_work_item_dimensions : int;
		address_bits : int;
		max_mem_alloc_size : int;
		image_support : bool;
		max_read_image_args : int;
		max_write_image_args : int;
		max_samplers : int;
		mem_base_addr_align : int;
		min_data_type_align_size : int;
		global_mem_cacheline_size : int;
		global_mem_cache_size : int;
		max_constant_args : int;
		endian_little : bool;
		available : bool;
		compiler_available : bool;
		single_fp_config : clDeviceFPConfig;
		global_mem_cache_type : clDeviceGlobalMemCacheType;
		queue_properties : clDeviceQueueProperties;
		local_mem_type : clDeviceLocalMemType;
		double_fp_config : clDeviceFPConfig;
		max_constant_buffer_size: int;
		execution_capabilities: clDeviceExecutionCapabilities;
		half_fp_config : clDeviceFPConfig;
		max_work_group_size : int;
		image2D_max_height : int;
		image2D_max_width : int;
		image3D_max_depth : int;
		image3D_max_height : int;
		image3D_max_width : int;
		max_parameter_size : int;
		max_work_item_size : dim3;
		prefered_vector_width_char : int;
		prefered_vector_width_short : int;
		prefered_vector_width_int : int;
		prefered_vector_width_long : int;
		prefered_vector_width_float : int;
		prefered_vector_width_double : int;
		profiling_timer_resolution : int;
		driver_version : string;
		device_id:clDeviceID;
	}	

type specificInfo =
	CudaInfo of cudaInfo
	| OpenCLInfo of openCLInfo
	| OtherInfo
	
	
	type device = {
	general_info : generalInfo;
	specific_info : specificInfo	
	}


external get_cuda_compatible_devices : unit -> int = "spoc_getCudaDevicesCount"
external get_opencl_compatible_devices : unit -> int = "spoc_getOpenCLDevicesCount"

external get_cuda_device : int -> device = "spoc_getCudaDevice"
external get_opencl_device : int -> int -> device = "spoc_getOpenCLDevice"

let cuda_compatible_devices = ref 0
(*get_cuda_compatible_devices()*)
let opencl_compatible_devices = ref 0
(*get_opencl_compatible_devices()*)

let total_num_devices = ref 0 
(*cuda_compatible_devices + opencl_compatible_devices*)

(*let devices = Array.init total_num_devices (*cuda_compatible_devices*)
		(fun i -> 
			if (i < cuda_compatible_devices) then
				get_cuda_device i 
			else 
				(*if (i < opencl_compatible_devices) then *)
					get_opencl_device (i-cuda_compatible_devices) i;

		) *)
		
let current_cuda_device = ref 0
let current_opencl_device = ref 0(*cuda_compatible_devices*)

(*let devices = ref [||]*)



(*let cudaKernelCache = ref [||]

let openCLKernelCache = ref [||]
let nbCaches = 2
*)

external is_available : int  -> bool = "spoc_opencl_is_available"

let init ?only:(s=Both) () = 
	begin
		match	s with
		| Both  -> (
			cuda_init ();
			cuda_compatible_devices := get_cuda_compatible_devices ();
			cl_init ();
			opencl_compatible_devices := get_opencl_compatible_devices();
			total_num_devices := !cuda_compatible_devices + !opencl_compatible_devices
			)
		| Cuda  -> (
			cuda_init ();
			cuda_compatible_devices := get_cuda_compatible_devices ();
			)		
		| OpenCL -> (
			cl_init ();
			opencl_compatible_devices := get_opencl_compatible_devices();
			)
	end;
	total_num_devices := !cuda_compatible_devices + !opencl_compatible_devices;
 	let devList = ref [] in
	for i = 0 to !cuda_compatible_devices - 1 do
			devList := !devList@[(get_cuda_device i)]
	done;
	let i = ref 0 and j = ref 0 in
	while !j < (!opencl_compatible_devices ) do
	if (is_available !i)  then
			(
				devList := !devList@[(get_opencl_device !i (!i + !cuda_compatible_devices))];
				incr i;
			);
		incr j;
	done;
	total_num_devices := List.length !devList;
	opencl_compatible_devices := !i;
	(*devices := Array.of_list	!devList;
	!devices*)
	Array.of_list	!devList
	
let	cuda_devices () = !cuda_compatible_devices
let	opencl_devices () = !opencl_compatible_devices
let	gpgpu_devices () = !total_num_devices

external cuda_flush : generalInfo -> unit = "spoc_cuda_flush"
external opencl_flush : generalInfo  -> int -> unit = "spoc_opencl_flush"


let flush dev ?queue_id:(q=(0)) () = 
	match dev.general_info.specific with
		| Cuda -> cuda_flush dev.general_info
		| OpenCL -> 
				opencl_flush dev.general_info q;
		| _  -> failwith "Flush Not implemented"


let hasCLExtension dev ext =
 match dev.specific_info with
		| OpenCLInfo cli -> 
			begin
				try
					ignore(Str.search_forward (Str.regexp ext) cli.extensions 0);
					true
				with
					| _ -> false
			end
		| _  -> false
